from datetime import datetime
from typing import List
from typing import Optional

from pendulum.constants import DAYS_PER_YEAR
from pendulum.constants import SECS_PER_YEAR
from pendulum.helpers import is_leap
from pendulum.helpers import local_time
from pendulum.helpers import timestamp
from pendulum.helpers import week_day

from .posix_timezone import PosixTimezone
from .transition import Transition
from .transition_type import TransitionType


class Timezone:
    def __init__(
        self,
        transitions,  # type: List[Transition]
        posix_rule=None,  # type: Optional[PosixTimezone]
        extended=True,  # type: bool
    ):
        self._posix_rule = posix_rule
        self._transitions = transitions

        if extended:
            self._extends()

    @property
    def transitions(self):  # type: () -> List[Transition]
        return self._transitions

    @property
    def posix_rule(self):
        return self._posix_rule

    def _extends(self):
        if not self._posix_rule:
            return

        posix = self._posix_rule

        if not posix.dst_abbr:
            # std only
            # The future specification should match the last/default transition
            ttype = self._transitions[-1].ttype
            if not self._check_ttype(ttype, posix.std_offset, False, posix.std_abbr):
                raise ValueError("Posix spec does not match last transition")

            return

        if len(self._transitions) < 2:
            raise ValueError("Too few transitions for POSIX spec")

        # Extend the transitions for an additional 400 years
        # using the future specification

        # The future specification should match the last two transitions,
        # and those transitions should have different is_dst flags.
        tr0 = self._transitions[-1]
        tr1 = self._transitions[-2]
        tt0 = tr0.ttype
        tt1 = tr1.ttype
        if tt0.is_dst():
            dst = tt0
            std = tt1
        else:
            dst = tt1
            std = tt0

        self._check_ttype(dst, posix.dst_offset, True, posix.dst_abbr)
        self._check_ttype(std, posix.std_offset, False, posix.std_abbr)

        # Add the transitions to tr1 and back to tr0 for each extra year.
        last_year = local_time(tr0.local, 0, 0)[0]
        leap_year = is_leap(last_year)
        jan1 = datetime(last_year, 1, 1)
        jan1_time = timestamp(jan1)
        jan1_weekday = week_day(jan1.year, jan1.month, jan1.day) % 7

        if local_time(tr1.local, 0, 0)[0] != last_year:
            # Add a single extra transition to align to a calendar year.
            if tt0.is_dst():
                pt1 = posix.dst_end
            else:
                pt1 = posix.dst_start

            tr1_offset = pt1.trans_offset(leap_year, jan1_weekday)
            tr = Transition(jan1_time + tr1_offset - tt0.offset, tr1.ttype, tr0)
            tr0 = tr
            tr1 = tr0
            tt0 = tr0.ttype
            tt1 = tr1.ttype

        if tt0.is_dst():
            pt1 = posix.dst_end
            pt0 = posix.dst_start
        else:
            pt1 = posix.dst_start
            pt0 = posix.dst_end

        tr = tr0
        for year in range(last_year + 1, last_year + 401):
            jan1_time += SECS_PER_YEAR[leap_year]
            jan1_weekday = (jan1_weekday + DAYS_PER_YEAR[leap_year]) % 7
            leap_year = not leap_year and is_leap(year)

            tr1_offset = pt1.trans_offset(leap_year, jan1_weekday)
            tr = Transition(jan1_time + tr1_offset - tt0.offset, tt1, tr)
            self._transitions.append(tr)

            tr0_offset = pt0.trans_offset(leap_year, jan1_weekday)
            tr = Transition(jan1_time + tr0_offset - tt1.offset, tt0, tr)
            self._transitions.append(tr)

    def _check_ttype(
        self,
        ttype,  # type: TransitionType
        offset,  # type: int
        is_dst,  # type: bool
        abbr,  # type: str
    ):  # type: (...) -> bool
        return (
            ttype.offset == offset
            and ttype.is_dst() == is_dst
            and ttype.abbreviation == abbr
        )



from ..formatting import Formatter


_formatter = Formatter()


class FormattableMixin(object):

    _formatter = _formatter

    def format(self, fmt, locale=None):
        """
        Formats the instance using the given format.

        :param fmt: The format to use
        :type fmt: str

        :param locale: The locale to use
        :type locale: str or None

        :rtype: str
        """
        return self._formatter.format(self, fmt, locale)

    def for_json(self):
        """
        Methods for automatic json serialization by simplejson

        :rtype: str
        """
        return str(self)

    def __format__(self, format_spec):
        if len(format_spec) > 0:
            if "%" in format_spec:
                return self.strftime(format_spec)

            return self.format(format_spec)

        return str(self)

    def __str__(self):
        return self.isoformat()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals


"""
lt custom locale file.
"""

translations = {
    # Relative time
    "units_relative": {
        "year": {
            "future": {
                "other": "{0} metų",
                "one": "{0} metų",
                "few": "{0} metų",
                "many": "{0} metų",
            },
            "past": {
                "other": "{0} metų",
                "one": "{0} metus",
                "few": "{0} metus",
                "many": "{0} metų",
            },
        },
        "month": {
            "future": {
                "other": "{0} mėnesių",
                "one": "{0} mėnesio",
                "few": "{0} mėnesių",
                "many": "{0} mėnesio",
            },
            "past": {
                "other": "{0} mėnesių",
                "one": "{0} mėnesį",
                "few": "{0} mėnesius",
                "many": "{0} mėnesio",
            },
        },
        "week": {
            "future": {
                "other": "{0} savaičių",
                "one": "{0} savaitės",
                "few": "{0} savaičių",
                "many": "{0} savaitės",
            },
            "past": {
                "other": "{0} savaičių",
                "one": "{0} savaitę",
                "few": "{0} savaites",
                "many": "{0} savaitės",
            },
        },
        "day": {
            "future": {
                "other": "{0} dienų",
                "one": "{0} dienos",
                "few": "{0} dienų",
                "many": "{0} dienos",
            },
            "past": {
                "other": "{0} dienų",
                "one": "{0} dieną",
                "few": "{0} dienas",
                "many": "{0} dienos",
            },
        },
        "hour": {
            "future": {
                "other": "{0} valandų",
                "one": "{0} valandos",
                "few": "{0} valandų",
                "many": "{0} valandos",
            },
            "past": {
                "other": "{0} valandų",
                "one": "{0} valandą",
                "few": "{0} valandas",
                "many": "{0} valandos",
            },
        },
        "minute": {
            "future": {
                "other": "{0} minučių",
                "one": "{0} minutės",
                "few": "{0} minučių",
                "many": "{0} minutės",
            },
            "past": {
                "other": "{0} minučių",
                "one": "{0} minutę",
                "few": "{0} minutes",
                "many": "{0} minutės",
            },
        },
        "second": {
            "future": {
                "other": "{0} sekundžių",
                "one": "{0} sekundės",
                "few": "{0} sekundžių",
                "many": "{0} sekundės",
            },
            "past": {
                "other": "{0} sekundžių",
                "one": "{0} sekundę",
                "few": "{0} sekundes",
                "many": "{0} sekundės",
            },
        },
    },
    "after": "po {0}",
    "before": "{0} nuo dabar",
    # Date formats
    "date_formats": {
        "LTS": "HH:mm:ss",
        "LT": "HH:mm",
        "LLLL": "YYYY [m.] MMMM D [d.], dddd, HH:mm [val.]",
        "LLL": "YYYY [m.] MMMM D [d.], HH:mm [val.]",
        "LL": "YYYY [m.] MMMM D [d.]",
        "L": "YYYY-MM-DD",
    },
}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from .custom import translations as custom_translations


"""
pl locale file.

It has been generated automatically and must not be modified directly.
"""


locale = {
    "plural": lambda n: "few"
    if (
        (
            (0 == 0 and ((0 == 0)))
            and ((n % 10) == (n % 10) and (((n % 10) >= 2 and (n % 10) <= 4)))
        )
        and (not ((n % 100) == (n % 100) and (((n % 100) >= 12 and (n % 100) <= 14))))
    )
    else "many"
    if (
        (
            (
                ((0 == 0 and ((0 == 0))) and (not (n == n and ((n == 1)))))
                and ((n % 10) == (n % 10) and (((n % 10) >= 0 and (n % 10) <= 1)))
            )
            or (
                (0 == 0 and ((0 == 0)))
                and ((n % 10) == (n % 10) and (((n % 10) >= 5 and (n % 10) <= 9)))
            )
        )
        or (
            (0 == 0 and ((0 == 0)))
            and ((n % 100) == (n % 100) and (((n % 100) >= 12 and (n % 100) <= 14)))
        )
    )
    else "one"
    if ((n == n and ((n == 1))) and (0 == 0 and ((0 == 0))))
    else "other",
    "ordinal": lambda n: "other",
    "translations": {
        "days": {
            "abbreviated": {
                0: "niedz.",
                1: "pon.",
                2: "wt.",
                3: "śr.",
                4: "czw.",
                5: "pt.",
                6: "sob.",
            },
            "narrow": {0: "n", 1: "p", 2: "w", 3: "ś", 4: "c", 5: "p", 6: "s"},
            "short": {
                0: "nie",
                1: "pon",
                2: "wto",
                3: "śro",
                4: "czw",
                5: "pią",
                6: "sob",
            },
            "wide": {
                0: "niedziela",
                1: "poniedziałek",
                2: "wtorek",
                3: "środa",
                4: "czwartek",
                5: "piątek",
                6: "sobota",
            },
        },
        "months": {
            "abbreviated": {
                1: "sty",
                2: "lut",
                3: "mar",
                4: "kwi",
                5: "maj",
                6: "cze",
                7: "lip",
                8: "sie",
                9: "wrz",
                10: "paź",
                11: "lis",
                12: "gru",
            },
            "narrow": {
                1: "s",
                2: "l",
                3: "m",
                4: "k",
                5: "m",
                6: "c",
                7: "l",
                8: "s",
                9: "w",
                10: "p",
                11: "l",
                12: "g",
            },
            "wide": {
                1: "stycznia",
                2: "lutego",
                3: "marca",
                4: "kwietnia",
                5: "maja",
                6: "czerwca",
                7: "lipca",
                8: "sierpnia",
                9: "września",
                10: "października",
                11: "listopada",
                12: "grudnia",
            },
        },
        "units": {
            "year": {
                "one": "{0} rok",
                "few": "{0} lata",
                "many": "{0} lat",
                "other": "{0} roku",
            },
            "month": {
                "one": "{0} miesiąc",
                "few": "{0} miesiące",
                "many": "{0} miesięcy",
                "other": "{0} miesiąca",
            },
            "week": {
                "one": "{0} tydzień",
                "few": "{0} tygodnie",
                "many": "{0} tygodni",
                "other": "{0} tygodnia",
            },
            "day": {
                "one": "{0} dzień",
                "few": "{0} dni",
                "many": "{0} dni",
                "other": "{0} dnia",
            },
            "hour": {
                "one": "{0} godzina",
                "few": "{0} godziny",
                "many": "{0} godzin",
                "other": "{0} godziny",
            },
            "minute": {
                "one": "{0} minuta",
                "few": "{0} minuty",
                "many": "{0} minut",
                "other": "{0} minuty",
            },
            "second": {
                "one": "{0} sekunda",
                "few": "{0} sekundy",
                "many": "{0} sekund",
                "other": "{0} sekundy",
            },
            "microsecond": {
                "one": "{0} mikrosekunda",
                "few": "{0} mikrosekundy",
                "many": "{0} mikrosekund",
                "other": "{0} mikrosekundy",
            },
        },
        "relative": {
            "year": {
                "future": {
                    "other": "za {0} roku",
                    "one": "za {0} rok",
                    "few": "za {0} lata",
                    "many": "za {0} lat",
                },
                "past": {
                    "other": "{0} roku temu",
                    "one": "{0} rok temu",
                    "few": "{0} lata temu",
                    "many": "{0} lat temu",
                },
            },
            "month": {
                "future": {
                    "other": "za {0} miesiąca",
                    "one": "za {0} miesiąc",
                    "few": "za {0} miesiące",
                    "many": "za {0} miesięcy",
                },
                "past": {
                    "other": "{0} miesiąca temu",
                    "one": "{0} miesiąc temu",
                    "few": "{0} miesiące temu",
                    "many": "{0} miesięcy temu",
                },
            },
            "week": {
                "future": {
                    "other": "za {0} tygodnia",
                    "one": "za {0} tydzień",
                    "few": "za {0} tygodnie",
                    "many": "za {0} tygodni",
                },
                "past": {
                    "other": "{0} tygodnia temu",
                    "one": "{0} tydzień temu",
                    "few": "{0} tygodnie temu",
                    "many": "{0} tygodni temu",
                },
            },
            "day": {
                "future": {
                    "other": "za {0} dnia",
                    "one": "za {0} dzień",
                    "few": "za {0} dni",
                    "many": "za {0} dni",
                },
                "past": {
                    "other": "{0} dnia temu",
                    "one": "{0} dzień temu",
                    "few": "{0} dni temu",
                    "many": "{0} dni temu",
                },
            },
            "hour": {
                "future": {
                    "other": "za {0} godziny",
                    "one": "za {0} godzinę",
                    "few": "za {0} godziny",
                    "many": "za {0} godzin",
                },
                "past": {
                    "other": "{0} godziny temu",
                    "one": "{0} godzinę temu",
                    "few": "{0} godziny temu",
                    "many": "{0} godzin temu",
                },
            },
            "minute": {
                "future": {
                    "other": "za {0} minuty",
                    "one": "za {0} minutę",
                    "few": "za {0} minuty",
                    "many": "za {0} minut",
                },
                "past": {
                    "other": "{0} minuty temu",
                    "one": "{0} minutę temu",
                    "few": "{0} minuty temu",
                    "many": "{0} minut temu",
                },
            },
            "second": {
                "future": {
                    "other": "za {0} sekundy",
                    "one": "za {0} sekundę",
                    "few": "za {0} sekundy",
                    "many": "za {0} sekund",
                },
                "past": {
                    "other": "{0} sekundy temu",
                    "one": "{0} sekundę temu",
                    "few": "{0} sekundy temu",
                    "many": "{0} sekund temu",
                },
            },
        },
        "day_periods": {
            "midnight": "o północy",
            "am": "AM",
            "noon": "w południe",
            "pm": "PM",
            "morning1": "rano",
            "morning2": "przed południem",
            "afternoon1": "po południu",
            "evening1": "wieczorem",
            "night1": "w nocy",
        },
    },
    "custom": custom_translations,
}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from .custom import translations as custom_translations


"""
fo locale file.

It has been generated automatically and must not be modified directly.
"""


locale = {
    "plural": lambda n: "one" if (n == n and ((n == 1))) else "other",
    "ordinal": lambda n: "other",
    "translations": {
        "days": {
            "abbreviated": {
                0: "sun.",
                1: "mán.",
                2: "týs.",
                3: "mik.",
                4: "hós.",
                5: "frí.",
                6: "ley.",
            },
            "narrow": {0: "S", 1: "M", 2: "T", 3: "M", 4: "H", 5: "F", 6: "L"},
            "short": {
                0: "su.",
                1: "má.",
                2: "tý.",
                3: "mi.",
                4: "hó.",
                5: "fr.",
                6: "le.",
            },
            "wide": {
                0: "sunnudagur",
                1: "mánadagur",
                2: "týsdagur",
                3: "mikudagur",
                4: "hósdagur",
                5: "fríggjadagur",
                6: "leygardagur",
            },
        },
        "months": {
            "abbreviated": {
                1: "jan.",
                2: "feb.",
                3: "mar.",
                4: "apr.",
                5: "mai",
                6: "jun.",
                7: "jul.",
                8: "aug.",
                9: "sep.",
                10: "okt.",
                11: "nov.",
                12: "des.",
            },
            "narrow": {
                1: "J",
                2: "F",
                3: "M",
                4: "A",
                5: "M",
                6: "J",
                7: "J",
                8: "A",
                9: "S",
                10: "O",
                11: "N",
                12: "D",
            },
            "wide": {
                1: "januar",
                2: "februar",
                3: "mars",
                4: "apríl",
                5: "mai",
                6: "juni",
                7: "juli",
                8: "august",
                9: "september",
                10: "oktober",
                11: "november",
                12: "desember",
            },
        },
        "units": {
            "year": {"one": "{0} ár", "other": "{0} ár"},
            "month": {"one": "{0} mánaður", "other": "{0} mánaðir"},
            "week": {"one": "{0} vika", "other": "{0} vikur"},
            "day": {"one": "{0} dagur", "other": "{0} dagar"},
            "hour": {"one": "{0} tími", "other": "{0} tímar"},
            "minute": {"one": "{0} minuttur", "other": "{0} minuttir"},
            "second": {"one": "{0} sekund", "other": "{0} sekundir"},
            "microsecond": {"one": "{0} mikrosekund", "other": "{0} mikrosekundir"},
        },
        "relative": {
            "year": {
                "future": {"other": "um {0} ár", "one": "um {0} ár"},
                "past": {"other": "{0} ár síðan", "one": "{0} ár síðan"},
            },
            "month": {
                "future": {"other": "um {0} mánaðir", "one": "um {0} mánað"},
                "past": {"other": "{0} mánaðir síðan", "one": "{0} mánað síðan"},
            },
            "week": {
                "future": {"other": "um {0} vikur", "one": "um {0} viku"},
                "past": {"other": "{0} vikur síðan", "one": "{0} vika síðan"},
            },
            "day": {
                "future": {"other": "um {0} dagar", "one": "um {0} dag"},
                "past": {"other": "{0} dagar síðan", "one": "{0} dagur síðan"},
            },
            "hour": {
                "future": {"other": "um {0} tímar", "one": "um {0} tíma"},
                "past": {"other": "{0} tímar síðan", "one": "{0} tími síðan"},
            },
            "minute": {
                "future": {"other": "um {0} minuttir", "one": "um {0} minutt"},
                "past": {"other": "{0} minuttir síðan", "one": "{0} minutt síðan"},
            },
            "second": {
                "future": {"other": "um {0} sekund", "one": "um {0} sekund"},
                "past": {"other": "{0} sekund síðan", "one": "{0} sekund síðan"},
            },
        },
        "day_periods": {"am": "AM", "pm": "PM"},
    },
    "custom": custom_translations,
}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals


"""
id custom locale file.
"""

translations = {
    "units": {"few_second": "beberapa detik"},
    "ago": "{} yang lalu",
    "from_now": "dalam {}",
    "after": "{0} kemudian",
    "before": "{0} yang lalu",
    "date_formats": {
        "LTS": "HH:mm:ss",
        "LT": "HH:mm",
        "LLLL": "dddd [d.] D. MMMM YYYY HH:mm",
        "LLL": "D. MMMM YYYY HH:mm",
        "LL": "D. MMMM YYYY",
        "L": "DD/MM/YYYY",
    },
}



from __future__ import absolute_import

import datetime
import typing

import pendulum

from .date import Date
from .datetime import DateTime
from .parsing import _Interval
from .parsing import parse as base_parse
from .time import Duration
from .time import Time
from .tz import UTC


try:
    from .parsing._iso8601 import Duration as CDuration
except ImportError:
    CDuration = None


def parse(
    text, **options
):  # type: (str, **typing.Any) -> typing.Union[Date, Time, DateTime, Duration]
    # Use the mock now value if it exists
    options["now"] = options.get("now", pendulum.get_test_now())

    return _parse(text, **options)


def _parse(text, **options):
    """
    Parses a string with the given options.

    :param text: The string to parse.
    :type text: str

    :rtype: mixed
    """
    # Handling special cases
    if text == "now":
        return pendulum.now()

    parsed = base_parse(text, **options)

    if isinstance(parsed, datetime.datetime):
        return pendulum.datetime(
            parsed.year,
            parsed.month,
            parsed.day,
            parsed.hour,
            parsed.minute,
            parsed.second,
            parsed.microsecond,
            tz=parsed.tzinfo or options.get("tz", UTC),
        )

    if isinstance(parsed, datetime.date):
        return pendulum.date(parsed.year, parsed.month, parsed.day)

    if isinstance(parsed, datetime.time):
        return pendulum.time(
            parsed.hour, parsed.minute, parsed.second, parsed.microsecond
        )

    if isinstance(parsed, _Interval):
        if parsed.duration is not None:
            duration = parsed.duration

            if parsed.start is not None:
                dt = pendulum.instance(parsed.start, tz=options.get("tz", UTC))

                return pendulum.period(
                    dt,
                    dt.add(
                        years=duration.years,
                        months=duration.months,
                        weeks=duration.weeks,
                        days=duration.remaining_days,
                        hours=duration.hours,
                        minutes=duration.minutes,
                        seconds=duration.remaining_seconds,
                        microseconds=duration.microseconds,
                    ),
                )

            dt = pendulum.instance(parsed.end, tz=options.get("tz", UTC))

            return pendulum.period(
                dt.subtract(
                    years=duration.years,
                    months=duration.months,
                    weeks=duration.weeks,
                    days=duration.remaining_days,
                    hours=duration.hours,
                    minutes=duration.minutes,
                    seconds=duration.remaining_seconds,
                    microseconds=duration.microseconds,
                ),
                dt,
            )

        return pendulum.period(
            pendulum.instance(parsed.start, tz=options.get("tz", UTC)),
            pendulum.instance(parsed.end, tz=options.get("tz", UTC)),
        )

    if CDuration and isinstance(parsed, CDuration):
        return pendulum.duration(
            years=parsed.years,
            months=parsed.months,
            weeks=parsed.weeks,
            days=parsed.days,
            hours=parsed.hours,
            minutes=parsed.minutes,
            seconds=parsed.seconds,
            microseconds=parsed.microseconds,
        )

    return parsed



# -*- coding: utf-8 -*-
from __future__ import unicode_literals


"""
fr custom locale file.
"""

translations = {
    "units": {"few_second": "quelques secondes"},
    # Relative Time
    "ago": "il y a {0}",
    "from_now": "dans {0}",
    "after": "{0} après",
    "before": "{0} avant",
    # Ordinals
    "ordinal": {"one": "er", "other": "e"},
    # Date formats
    "date_formats": {
        "LTS": "HH:mm:ss",
        "LT": "HH:mm",
        "LLLL": "dddd D MMMM YYYY HH:mm",
        "LLL": "D MMMM YYYY HH:mm",
        "LL": "D MMMM YYYY",
        "L": "DD/MM/YYYY",
    },
}



from __future__ import absolute_import

import datetime as _datetime

from typing import Optional
from typing import Union

from .__version__ import __version__
from .constants import DAYS_PER_WEEK
from .constants import FRIDAY
from .constants import HOURS_PER_DAY
from .constants import MINUTES_PER_HOUR
from .constants import MONDAY
from .constants import MONTHS_PER_YEAR
from .constants import SATURDAY
from .constants import SECONDS_PER_DAY
from .constants import SECONDS_PER_HOUR
from .constants import SECONDS_PER_MINUTE
from .constants import SUNDAY
from .constants import THURSDAY
from .constants import TUESDAY
from .constants import WEDNESDAY
from .constants import WEEKS_PER_YEAR
from .constants import YEARS_PER_CENTURY
from .constants import YEARS_PER_DECADE
from .date import Date
from .datetime import DateTime
from .duration import Duration
from .formatting import Formatter
from .helpers import format_diff
from .helpers import get_locale
from .helpers import get_test_now
from .helpers import has_test_now
from .helpers import locale
from .helpers import set_locale
from .helpers import set_test_now
from .helpers import test
from .helpers import week_ends_at
from .helpers import week_starts_at
from .parser import parse
from .period import Period
from .time import Time
from .tz import POST_TRANSITION
from .tz import PRE_TRANSITION
from .tz import TRANSITION_ERROR
from .tz import UTC
from .tz import local_timezone
from .tz import set_local_timezone
from .tz import test_local_timezone
from .tz import timezone
from .tz import timezones
from .tz.timezone import Timezone as _Timezone
from .utils._compat import _HAS_FOLD


_TEST_NOW = None  # type: Optional[DateTime]
_LOCALE = "en"
_WEEK_STARTS_AT = MONDAY
_WEEK_ENDS_AT = SUNDAY

_formatter = Formatter()


def _safe_timezone(obj):
    # type: (Optional[Union[str, float, _datetime.tzinfo, _Timezone]]) -> _Timezone
    """
    Creates a timezone instance
    from a string, Timezone, TimezoneInfo or integer offset.
    """
    if isinstance(obj, _Timezone):
        return obj

    if obj is None or obj == "local":
        return local_timezone()

    if isinstance(obj, (int, float)):
        obj = int(obj * 60 * 60)
    elif isinstance(obj, _datetime.tzinfo):
        # pytz
        if hasattr(obj, "localize"):
            obj = obj.zone
        elif obj.tzname(None) == "UTC":
            return UTC
        else:
            offset = obj.utcoffset(None)

            if offset is None:
                offset = _datetime.timedelta(0)

            obj = int(offset.total_seconds())

    return timezone(obj)


# Public API
def datetime(
    year,  # type: int
    month,  # type: int
    day,  # type: int
    hour=0,  # type: int
    minute=0,  # type: int
    second=0,  # type: int
    microsecond=0,  # type: int
    tz=UTC,  # type: Optional[Union[str, float, _Timezone]]
    dst_rule=POST_TRANSITION,  # type: str
):  # type: (...) -> DateTime
    """
    Creates a new DateTime instance from a specific date and time.
    """
    if tz is not None:
        tz = _safe_timezone(tz)

    if not _HAS_FOLD:
        dt = naive(year, month, day, hour, minute, second, microsecond)
    else:
        dt = _datetime.datetime(year, month, day, hour, minute, second, microsecond)
    if tz is not None:
        dt = tz.convert(dt, dst_rule=dst_rule)

    return DateTime(
        dt.year,
        dt.month,
        dt.day,
        dt.hour,
        dt.minute,
        dt.second,
        dt.microsecond,
        tzinfo=dt.tzinfo,
        fold=dt.fold,
    )


def local(
    year, month, day, hour=0, minute=0, second=0, microsecond=0
):  # type: (int, int, int, int, int, int, int) -> DateTime
    """
    Return a DateTime in the local timezone.
    """
    return datetime(
        year, month, day, hour, minute, second, microsecond, tz=local_timezone()
    )


def naive(
    year, month, day, hour=0, minute=0, second=0, microsecond=0
):  # type: (int, int, int, int, int, int, int) -> DateTime
    """
    Return a naive DateTime.
    """
    return DateTime(year, month, day, hour, minute, second, microsecond)


def date(year, month, day):  # type: (int, int, int) -> Date
    """
    Create a new Date instance.
    """
    return Date(year, month, day)


def time(hour, minute=0, second=0, microsecond=0):  # type: (int, int, int, int) -> Time
    """
    Create a new Time instance.
    """
    return Time(hour, minute, second, microsecond)


def instance(
    dt, tz=UTC
):  # type: (_datetime.datetime, Optional[Union[str, _Timezone]]) -> DateTime
    """
    Create a DateTime instance from a datetime one.
    """
    if not isinstance(dt, _datetime.datetime):
        raise ValueError("instance() only accepts datetime objects.")

    if isinstance(dt, DateTime):
        return dt

    tz = dt.tzinfo or tz

    # Checking for pytz/tzinfo
    if isinstance(tz, _datetime.tzinfo) and not isinstance(tz, _Timezone):
        # pytz
        if hasattr(tz, "localize") and tz.zone:
            tz = tz.zone
        else:
            # We have no sure way to figure out
            # the timezone name, we fallback
            # on a fixed offset
            tz = tz.utcoffset(dt).total_seconds() / 3600

    return datetime(
        dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond, tz=tz
    )


def now(tz=None):  # type: (Optional[Union[str, _Timezone]]) -> DateTime
    """
    Get a DateTime instance for the current date and time.
    """
    if has_test_now():
        test_instance = get_test_now()
        _tz = _safe_timezone(tz)

        if tz is not None and _tz != test_instance.timezone:
            test_instance = test_instance.in_tz(_tz)

        return test_instance

    if tz is None or tz == "local":
        dt = _datetime.datetime.now(local_timezone())
    elif tz is UTC or tz == "UTC":
        dt = _datetime.datetime.now(UTC)
    else:
        dt = _datetime.datetime.now(UTC)
        tz = _safe_timezone(tz)
        dt = tz.convert(dt)

    return DateTime(
        dt.year,
        dt.month,
        dt.day,
        dt.hour,
        dt.minute,
        dt.second,
        dt.microsecond,
        tzinfo=dt.tzinfo,
        fold=dt.fold if _HAS_FOLD else 0,
    )


def today(tz="local"):  # type: (Union[str, _Timezone]) -> DateTime
    """
    Create a DateTime instance for today.
    """
    return now(tz).start_of("day")


def tomorrow(tz="local"):  # type: (Union[str, _Timezone]) -> DateTime
    """
    Create a DateTime instance for today.
    """
    return today(tz).add(days=1)


def yesterday(tz="local"):  # type: (Union[str, _Timezone]) -> DateTime
    """
    Create a DateTime instance for today.
    """
    return today(tz).subtract(days=1)


def from_format(
    string, fmt, tz=UTC, locale=None,  # noqa
):  # type: (str, str, Union[str, _Timezone], Optional[str]) -> DateTime
    """
    Creates a DateTime instance from a specific format.
    """
    parts = _formatter.parse(string, fmt, now(), locale=locale)
    if parts["tz"] is None:
        parts["tz"] = tz

    return datetime(**parts)


def from_timestamp(
    timestamp, tz=UTC
):  # type: (Union[int, float], Union[str, _Timezone]) -> DateTime
    """
    Create a DateTime instance from a timestamp.
    """
    dt = _datetime.datetime.utcfromtimestamp(timestamp)

    dt = datetime(
        dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond
    )

    if tz is not UTC or tz != "UTC":
        dt = dt.in_timezone(tz)

    return dt


def duration(
    days=0,  # type: float
    seconds=0,  # type: float
    microseconds=0,  # type: float
    milliseconds=0,  # type: float
    minutes=0,  # type: float
    hours=0,  # type: float
    weeks=0,  # type: float
    years=0,  # type: float
    months=0,  # type: float
):  # type: (...) -> Duration
    """
    Create a Duration instance.
    """
    return Duration(
        days=days,
        seconds=seconds,
        microseconds=microseconds,
        milliseconds=milliseconds,
        minutes=minutes,
        hours=hours,
        weeks=weeks,
        years=years,
        months=months,
    )


def period(start, end, absolute=False):  # type: (DateTime, DateTime, bool) -> Period
    """
    Create a Period instance.
    """
    return Period(start, end, absolute=absolute)



import os

from collections import namedtuple
from struct import unpack
from typing import IO
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import pytzdata

from pytzdata.exceptions import TimezoneNotFound

from pendulum.utils._compat import PY2

from .exceptions import InvalidTimezone
from .exceptions import InvalidZoneinfoFile
from .posix_timezone import PosixTimezone
from .posix_timezone import posix_spec
from .timezone import Timezone
from .transition import Transition
from .transition_type import TransitionType


_offset = namedtuple("offset", "utc_total_offset is_dst abbr_idx")

header = namedtuple(
    "header",
    "version " "utclocals " "stdwalls " "leaps " "transitions " "types " "abbr_size",
)


class Reader:
    """
    Reads compiled zoneinfo TZif (\0, 2 or 3) files.
    """

    def __init__(self, extend=True):  # type: (bool) -> None
        self._extend = extend

    def read_for(self, timezone):  # type: (str) -> Timezone
        """
        Read the zoneinfo structure for a given timezone name.

        :param timezone: The timezone.
        """
        try:
            file_path = pytzdata.tz_path(timezone)
        except TimezoneNotFound:
            raise InvalidTimezone(timezone)

        return self.read(file_path)

    def read(self, file_path):  # type: (str) -> Timezone
        """
        Read a zoneinfo structure from the given path.

        :param file_path: The path of a zoneinfo file.
        """
        if not os.path.exists(file_path):
            raise InvalidZoneinfoFile("The tzinfo file does not exist")

        with open(file_path, "rb") as fd:
            return self._parse(fd)

    def _check_read(self, fd, nbytes):  # type: (...) -> bytes
        """
        Reads the given number of bytes from the given file
        and checks that the correct number of bytes could be read.
        """
        result = fd.read(nbytes)

        if (not result and nbytes > 0) or len(result) != nbytes:
            raise InvalidZoneinfoFile(
                "Expected {} bytes reading {}, "
                "but got {}".format(nbytes, fd.name, len(result) if result else 0)
            )

        if PY2:
            return bytearray(result)

        return result

    def _parse(self, fd):  # type: (...) -> Timezone
        """
        Parse a zoneinfo file.
        """
        hdr = self._parse_header(fd)

        if hdr.version in (2, 3):
            # We're skipping the entire v1 file since
            # at least the same data will be found in TZFile 2.
            fd.seek(
                hdr.transitions * 5
                + hdr.types * 6
                + hdr.abbr_size
                + hdr.leaps * 4
                + hdr.stdwalls
                + hdr.utclocals,
                1,
            )

            # Parse the second header
            hdr = self._parse_header(fd)

            if hdr.version != 2 and hdr.version != 3:
                raise InvalidZoneinfoFile(
                    "Header versions mismatch for file {}".format(fd.name)
                )

            # Parse the v2 data
            trans = self._parse_trans_64(fd, hdr.transitions)
            type_idx = self._parse_type_idx(fd, hdr.transitions)
            types = self._parse_types(fd, hdr.types)
            abbrs = self._parse_abbrs(fd, hdr.abbr_size, types)

            fd.seek(hdr.leaps * 8 + hdr.stdwalls + hdr.utclocals, 1)

            trule = self._parse_posix_tz(fd)
        else:
            # TZFile v1
            trans = self._parse_trans_32(fd, hdr.transitions)
            type_idx = self._parse_type_idx(fd, hdr.transitions)
            types = self._parse_types(fd, hdr.types)
            abbrs = self._parse_abbrs(fd, hdr.abbr_size, types)
            trule = None

        types = [
            TransitionType(off, is_dst, abbrs[abbr]) for off, is_dst, abbr in types
        ]

        transitions = []
        previous = None
        for trans, idx in zip(trans, type_idx):
            transition = Transition(trans, types[idx], previous)
            transitions.append(transition)

            previous = transition

        if not transitions:
            transitions.append(Transition(0, types[0], None))

        return Timezone(transitions, posix_rule=trule, extended=self._extend)

    def _parse_header(self, fd):  # type: (...) -> header
        buff = self._check_read(fd, 44)

        if buff[:4] != b"TZif":
            raise InvalidZoneinfoFile(
                'The file "{}" has an invalid header.'.format(fd.name)
            )

        version = {0x00: 1, 0x32: 2, 0x33: 3}.get(buff[4])

        if version is None:
            raise InvalidZoneinfoFile(
                'The file "{}" has an invalid version.'.format(fd.name)
            )

        hdr = header(version, *unpack(">6l", buff[20:44]))

        return hdr

    def _parse_trans_64(self, fd, n):  # type: (IO[Any], int) -> List[int]
        trans = []
        for _ in range(n):
            buff = self._check_read(fd, 8)
            trans.append(unpack(">q", buff)[0])

        return trans

    def _parse_trans_32(self, fd, n):  # type: (IO[Any], int) -> List[int]
        trans = []
        for _ in range(n):
            buff = self._check_read(fd, 4)
            trans.append(unpack(">i", buff)[0])

        return trans

    def _parse_type_idx(self, fd, n):  # type: (IO[Any], int) -> List[int]
        buff = self._check_read(fd, n)

        return list(unpack("{}B".format(n), buff))

    def _parse_types(
        self, fd, n
    ):  # type: (IO[Any], int) -> List[Tuple[Any, bool, int]]
        types = []

        for _ in range(n):
            buff = self._check_read(fd, 6)
            offset = unpack(">l", buff[:4])[0]
            is_dst = buff[4] == 1
            types.append((offset, is_dst, buff[5]))

        return types

    def _parse_abbrs(
        self, fd, n, types
    ):  # type: (IO[Any], int, List[Tuple[Any, bool, int]]) -> Dict[int, str]
        abbrs = {}
        buff = self._check_read(fd, n)

        for offset, is_dst, idx in types:
            if idx not in abbrs:
                abbr = buff[idx : buff.find(b"\0", idx)].decode("utf-8")
                abbrs[idx] = abbr

        return abbrs

    def _parse_posix_tz(self, fd):  # type: (...) -> Optional[PosixTimezone]
        s = fd.read().decode("utf-8")

        if not s.startswith("\n") or not s.endswith("\n"):
            raise InvalidZoneinfoFile('Invalid posix rule in file "{}"'.format(fd.name))

        s = s.strip()

        if not s:
            return

        return posix_spec(s)



# -*- coding: utf-8 -*-
"""
it custom locale file.
"""

from __future__ import unicode_literals


translations = {
    "units": {"few_second": "alcuni secondi"},
    # Relative Time
    "ago": "{0} fa",
    "from_now": "in {0}",
    "after": "{0} dopo",
    "before": "{0} prima",
    # Ordinals
    "ordinal": {"other": "°"},
    # Date formats
    "date_formats": {
        "LTS": "H:mm:ss",
        "LT": "H:mm",
        "L": "DD/MM/YYYY",
        "LL": "D MMMM YYYY",
        "LLL": "D MMMM YYYY [alle] H:mm",
        "LLLL": "dddd, D MMMM YYYY [alle] H:mm",
    },
}
